/**********************************************************************
    Copyright 2020, Batakrishna Sahu, mailto:<Batakrishna.Sahu@suiit.ac.in> find license text at end of file
/**********************************************************************

/* Override Nuitka logo size */
#projectlogo img {
    max-width: 50px;
    padding: 5px;
}

/**********************************************************************
    Part of "Nuitka", an optimizing Python compiler that is compatible and
    integrates with CPython, but also works on its own.

    Licensed under the GNU Affero General Public License, Version 3 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       https://www.gnu.org/licenses/agpl-3.0.txt

    See also: "Nuitka Runtime Library Exception, Version 1.0" in file
    "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
/**********************************************************************
