{#-  Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file -#}

// This file is automatically generated by misc/vscode_config_gen.py
// DO NOT EDIT THIS FILE DIRECTLY.
// Edit .vscode/c_cpp_properties.json.j2 or misc/vscode_config_gen.py instead.
{
    "env": {
        "myDefaultIncludePath": [
            "${workspaceFolder}/nuitka/build/include",
            "${workspaceFolder}/nuitka/build/static_src",
            "${workspaceFolder}/nuitka/build/inline_copy/zstd",
            "${workspaceFolder}/nuitka/build/inline_copy/zlib",
            "${workspaceFolder}/nuitka/build/inline_copy/libbacktrace"{% if plugin_include_paths %},
{% endif %}
            {% for path in plugin_include_paths %}
            "{{ path }}"{{ "," if not loop.last else "" }}
            {% endfor %}
        ]
    },
    "configurations": [
        {
            "name": "Default",
            "includePath": [
                "${myDefaultIncludePath}",
                "{{ python_include_path | replace('\\', '\\\\') }}"
            ],
            "defines": [
                "_DEBUG",
                "UNICODE",
                "_UNICODE",
                "__IDE_ONLY__",
                "PYTHON_VERSION={{ python_version_hex }}"
            ],
            {% if windows_sdk_version %}
            "windowsSdkVersion": "{{ windows_sdk_version }}",
            {% endif %}
            "compilerPath": "{{ compiler_path | replace('\\', '/') }}",
            "cStandard": "c11",
            "cppStandard": "c++17",
            "intelliSenseMode": "{{ intelliSenseMode }}"
        }
    ],
    "version": 4
}

{#-  Part of "Nuitka", an optimizing Python compiler that is compatible and           -#}
{#-  integrates with CPython, but also works on its own.                              -#}
{#-                                                                                   -#}
{#-  Licensed under the GNU Affero General Public License, Version 3 (the "License"); -#}
{#-  you may not use this file except in compliance with the License.                 -#}
{#-  You may obtain a copy of the License at                                          -#}
{#-                                                                                   -#}
{#-  https://www.gnu.org/licenses/agpl-3.0.txt                                        -#}
{#-                                                                                   -#}
{#-  See also: "Nuitka Runtime Library Exception, Version 1.0" in file                -#}
{#-  "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.        -#}
{#-                                                                                   -#}
{#-  Unless required by applicable law or agreed to in writing, software              -#}
{#-  distributed under the License is distributed on an "AS IS" BASIS,                -#}
{#-  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.         -#}
{#-  See the License for the specific language governing permissions and              -#}
{#-  limitations under the License.                                                   -#}
